@@ -7,13 +7,15 @@ __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 DataStore::CAS::FileCreatorHandle - Handle-emulation object used for writing content into a CAS
 
 =head1 VERSION
 
-version 0.020001
+version 0.03
 
 =head1 DESCRIPTION
 
@@ -60,13 +62,25 @@ objects.
 
   Passes through to C<$cas-E<gt>_handle_close($handle, ...)>
 
+=head2 read
+
+Always returns 0
+
+=head2 readline
+
+Always returns C<undef>.
+
+=head2 eof
+
+Always returns 1 (handle cannot be read)
+
 =head1 AUTHOR
 
 Michael Conrad <mconrad@intellitree.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2013 by Michael Conrad, and IntelliTree Solutions llc.
+This software is copyright (c) 2022 by Michael Conrad, and IntelliTree Solutions llc.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
