@@ -8,7 +8,8 @@ use File::Spec 3.33;
 use File::Spec::Functions 'catfile', 'catdir', 'canonpath';
 use File::Temp 0.22 ();
 
-our $VERSION = '0.020001';
+our $VERSION = '0.03';
+our @CARP_NOT= qw( DataStore::CAS DataStore::CAS::File DataStore::CAS::VirtualHandle );
 
 # ABSTRACT: Simple file/directory based CAS implementation
 
@@ -19,14 +20,8 @@ has _config          => ( is => 'rwp', init_arg => undef );
 sub fanout              { [ $_[0]->fanout_list ] }
 sub fanout_list         { @{ $_[0]->_config->{fanout} } }
 sub digest              { $_[0]->_config->{digest} }
-
-has _fanout_regex    => ( is => 'lazy' );
-
-sub _build__fanout_regex {
-	my $self= shift;
-	my $regex= join('', map { "(.{$_})" } $self->fanout_list ).'(.*)';
-	qr/$regex/;
-}
+has _digest_hash_to_hex => ( is => 'rw', init_arg => undef );
+has _digest_hash_split  => ( is => 'rw', init_arg => undef );
 
 with 'DataStore::CAS';
 
@@ -62,6 +57,9 @@ sub BUILD {
 	}
 
 	$self->_set__config( $self->_load_config($path, { ignore_version => $ignore_version }) );
+	my ($tohex, $split)= _get_hex_and_fanout_functions($self->digest, $self->fanout);
+	$self->_digest_hash_to_hex($tohex);
+	$self->_digest_hash_split($split);
 
 	if ($setup) {
 		$self->put('');
@@ -76,6 +74,28 @@ sub BUILD {
 }
 
 
+sub path_parts_for_hash {
+	my ($self, $hash)= @_;
+	$self->_digest_hash_split->($hash);
+}
+
+sub path_for_hash {
+	my ($self, $hash, $create_dirs)= @_;
+	my @parts= $self->_digest_hash_split->($hash);
+	if ($create_dirs) {
+		my $path= $self->path;
+		for (@parts[0..($#parts-1)]) {
+			$path= catdir($path, $_);
+			next if -d $path;
+			mkdir($path) or croak "mkdir($path): $!";
+		}
+		return catfile($path, $parts[-1]);
+	} else {
+		return catfile($self->path, @parts);
+	}
+}
+
+
 sub create_store {
 	my $class= shift;
 	$class= ref $class if ref $class;
@@ -88,10 +108,7 @@ sub create_store {
 		unless $class->_is_dir_empty($params{path});
 
 	$params{digest} ||= 'SHA-1';
-	# Make sure the algorithm is available
-	my $found= ( try { defined $class->_new_digest($params{digest}); } catch { print "#$_\n"; 0; } )
-		or croak "Digest algorithm '".$params{digest}."'"
-		        ." is not available on this system.\n";
+	$class->_assert_digest_available($params{digest});
 
 	$params{fanout} ||= [ 1, 2 ];
 	# make sure the fanout isn't insane
@@ -130,18 +147,64 @@ sub _load_config {
 	# Get the digest algorithm name
 	$params{digest}=
 		$class->_parse_digest($class->_read_config_setting($path, 'digest'));
-	# Check for digest algorithm availability
-	my $found= ( try { $class->_new_digest($params{digest}); 1; } catch { 0; } )
-		or croak "Digest algorithm '".$params{digest}."'"
-		        ." is not available on this system.\n";
-
+	$class->_assert_digest_available($params{digest});
 	# Get the directory fan-out specification
-	$params{fanout}=
-		$class->_parse_fanout($class->_read_config_setting($path, 'fanout'));
-
+	$params{fanout}= $class->_parse_fanout($class->_read_config_setting($path, 'fanout'));
 	return \%params;
 }
 
+sub _get_hex_and_fanout_functions {
+	my ($digest, $fanout)= @_;
+	my $hexlen= length Digest->new($digest)->add('')->hexdigest;
+	my $rawlen= length Digest->new($digest)->add('')->digest;
+	# Create a function that coerces the argument into a hex string, or dies.
+	# When given a digest, it can be raw bytes, or hex.  The  hex one is double the length.
+	my $tohex= sub {
+		my $hash= $_[2];
+		my $len= length($hash) || 0;
+		$len == $hexlen? $hash
+		: $len == $rawlen? _to_hex($hash)
+		: croak "Invalid length for checksum of $digest: $len "._quoted($hash);
+	};
+
+	# Create a function that splits a digest into the path components
+	# for the CAS file.
+	$fanout= [ @$fanout ];
+	# final component might be a character indicating full-name or remainder-name
+	my $filename_type= $fanout->[-1] =~ /^[0-9]+$/? '*'
+		: pop @$fanout;
+	my $re= '^'.join('', map "([0-9a-f]{$_})", map /([0-9]+)/, @$fanout);
+	$re .= '([0-9a-f]+)' if $filename_type eq '*';
+	$re = qr/$re/;
+	my $split= ($filename_type eq '=')? sub {
+			my $hash= $_[0];
+			$hash= $tohex->($hash) if $hexlen != (length($hash) || 0);
+			my @dirs= ($hash =~ $re) or croak "can't split hash '$hash' into requested fanout";
+			return @dirs, $hash;
+		}
+		: ($filename_type eq '*')? sub {
+			my $hash= $_[0];
+			$hash= $tohex->($hash) if $hexlen != (length($hash) || 0);
+			my @dirs= ($hash =~ $re) or croak "can't split hash '$hash' into requested fanout";
+			return @dirs;
+		}
+		: croak "Unrecognized filename indicator in fanout specification: '$filename_type'";
+
+	return ($tohex, $split);
+}
+
+sub _to_hex {
+	my $tmp= shift;
+	$tmp =~ s/./ sprintf("%02X", $_) /ge;
+	$tmp;
+}
+sub _quoted {
+	my $tmp= shift;
+	return "(undef)" unless defined $tmp;
+	$tmp =~ s/[\0-\x1F\x7F]/ sprintf("\\x%02X", $_) /ge;
+	qq{"$tmp"};
+}
+
 sub _is_dir_empty {
 	my (undef, $path)= @_;
 	opendir(my $dh, $path)
@@ -175,19 +238,29 @@ sub _read_config_setting {
 	return $str;
 }
 
+# 4 hex digits makes 65536 subdirectories in a single parent
+our $max_sane_level_fanout= 4;
+# 6 hex digits creates 16 million directories, more than that is probably a mistake
+our $max_sane_total_fanout= 6;
 sub _parse_fanout {
 	my (undef, $fanout)= @_;
 	chomp($fanout);
-	my @fanout;
+	my @fanout= split /\s+/, $fanout;
 	# Sanity check on the fanout
 	my $total_digits= 0;
-	for (split /\s+/, $fanout) {
-		($_ =~ /^(\d+)$/) or croak "Invalid fanout spec";
-		push @fanout, $1;
-		$total_digits+= $1;
-		croak "Too large fanout in one directory ($1)" if $1 > 3;
+	for (@fanout) {
+		if ($_ =~ /^(\d+)$/) {
+			$total_digits+= $1;
+			croak "Too large fanout in one directory ($1)" if $1 > $max_sane_level_fanout;
+		} elsif ($_ eq '=' or $_ eq '*') {
+			# indicates full hash for filename, or partial hash
+			# must be the final element
+			\$_ == \$fanout[-1] or croak "Fanout '+' or '=' can only be final element";
+		} else {
+			croak "Invalid character in fanout specification: '$_'";
+		}
 	}
-	croak "Too many digits of fanout! ($total_digits)" if $total_digits > 5;
+	croak "Too many digits of fanout! ($total_digits)" if $total_digits > $max_sane_total_fanout;
 	return \@fanout;
 }
 
@@ -213,7 +286,7 @@ sub _parse_version {
 
 sub get {
 	my ($self, $hash)= @_;
-	my $fname= catfile($self->_path_for_hash($hash));
+	my $fname= $self->path_for_hash($hash);
 	return undef
 		unless (my ($size, $blksize)= (stat $fname)[7,11]);
 	return bless {
@@ -228,13 +301,113 @@ sub get {
 }
 
 
+sub put_file {
+	my ($self, $file, $flags)= @_;
+	my $is_cas_file= ref $file && ref($file)->isa('DataStore::CAS::File');
+	
+	# Can only optimize if source is a real file
+	if ($flags->{hardlink} || ($flags->{move} && !$is_cas_file)) {
+		my $fname= 
+			($is_cas_file && $file->can('local_file') and length $file->local_file)? $file->local_file
+			: (ref $file && ref($file)->isa('Path::Class::File'))? "$file"
+			: (!ref $file)? $file
+			: undef;
+		if ($fname && -f $fname) {
+			my %known_hashes= $flags->{known_hashes}? %{$flags->{known_hashes}} : ();
+			# Apply reuse_hash feature, if requested
+			$known_hashes{$file->store->digest}= $file->hash
+				if $is_cas_file && $flags->{reuse_hash};
+			# Calculate the hash if it wasn't given.
+			my $hash= ($known_hashes{$self->digest} ||= $self->calculate_file_hash($fname));
+			# Have it already?
+			if (-f $self->path_for_hash($hash)) {
+				$flags->{stats}{dup_file_count}++
+					if $flags->{stats};
+				$self->_unlink_source_file($file, $flags)
+					if $flags->{move};
+				return $hash;
+			}
+			# Save hash for next step
+			$flags= { %$flags, known_hashes => \%known_hashes };
+			# Try the move or hardlink operation.  If it fails, it returns false,
+			# and this falls through to the default implementation that copies the
+			# file.
+			return $hash if $flags->{move}
+				? $self->_try_put_move($fname, $flags)
+				: $self->_try_put_hardlink($fname, $flags);
+		}
+	}
+	# Else use the default implementation which opens and reads the file.
+	return DataStore::CAS::put_file($self, $file, $flags);
+}
+
+sub _try_put_move {
+	my ($self, $file, $flags)= @_;
+	my $hash= $flags->{known_hashes}{$self->digest}; # calculated above
+	# Need to be on same filesystem for this to work.
+	my $dest= $self->path_for_hash($hash,1);
+	my $tmp= "$dest.tmp";
+	return 0 unless rename($file, $tmp);
+	# Need to be able to change ownership to current user and remove write bits.
+	try {
+		my ($mode, $uid, $gid)= (stat $tmp)[2,4,5]
+			or die "stat($tmp): $!\n";
+		if (!$flags->{dry_run}) {
+			chown($>, $), $tmp) or die "chown($> $), $tmp): $!\n"
+				if ($uid && $uid != $>) or ($gid && $gid != $) );
+			chmod(0444, $tmp) or die "chmod(0444, $tmp): $!\n"
+				if 0444 != ($mode & 0777);
+			rename($tmp, $dest)
+				or die "rename($tmp, $dest): $!\n";
+		}
+		# record that we added a new hash, if stats enabled.
+		if ($flags->{stats}) {
+			$flags->{stats}{new_file_count}++;
+			push @{ $flags->{stats}{new_files} ||= [] }, $hash;
+		}
+		$hash;
+	}
+	catch {
+		warn "Can't optimize CAS insertion with move: $_";
+		unlink $tmp;
+		0;
+	};
+}
+
+sub _try_put_hardlink {
+	my ($self, $file, $flags)= @_;
+	my $hash= $flags->{known_hashes}{$self->digest}; # calculated above
+	# Refuse to link a file that is writeable by anyone.
+	my ($mode, $uid, $gid)= (stat $file)[2,4,5];
+	defined $mode && !($mode & 0222)
+		or return 0;
+	# Refuse to link a file owned by anyone else other than root
+	(!$uid || $uid == $>) and (!$gid || $gid == $))
+		or return 0;
+	# looks ok.
+	my $dest= $self->path_for_hash($hash,1);
+	$flags->{dry_run}
+		or link($file, $dest)
+		or return 0;
+	# record that we added a new hash, if stats enabled.
+	if ($flags->{stats}) {
+		$flags->{stats}{new_file_count}++;
+		push @{ $flags->{stats}{new_files} ||= [] }, $hash;
+	}
+	# it worked
+	return $hash;
+}
+
+
 sub new_write_handle {
 	my ($self, $flags)= @_;
 	$flags ||= {};
+	my $known_hash= $flags->{known_hashes} && $flags->{known_hashes}{$self->digest};
+	$known_hash= undef unless defined $known_hash && length $known_hash;
 	my $data= {
 		wrote   => 0,
 		dry_run => $flags->{dry_run},
-		hash    => $flags->{known_digests}{$self->digest},
+		hash    => $known_hash,
 		stats   => $flags->{stats},
 	};
 	
@@ -242,7 +415,7 @@ sub new_write_handle {
 		unless $data->{dry_run};
 	
 	$data->{digest}= $self->_new_digest
-		unless defined $data->{known_hash};
+		unless defined $data->{hash};
 	
 	return DataStore::CAS::FileCreatorHandle->new($self, $data);
 }
@@ -297,7 +470,7 @@ sub commit_write_handle {
 sub _commit_file {
 	my ($self, $source_file, $hash, $flags)= @_;
 	# Find the destination file name
-	my $dest_name= $self->_path_for_hash($hash);
+	my $dest_name= $self->path_for_hash($hash);
 	# Only if we don't have it yet...
 	if (-f $dest_name) {
 		if ($flags->{stats}) {
@@ -305,12 +478,15 @@ sub _commit_file {
 		}
 	}
 	else {
-		# link it into place
-		# we check for missing directories after the first failure,
+		# make it read-only
+		chmod(0444, "$source_file") or croak "chmod(0444, $source_file): $!";
+		
+		# Rename it into place
+		# Check for missing directories after the first failure,
 		#   in the spirit of keeping the common case fast.
 		$flags->{dry_run}
-			or link($source_file, $dest_name)
-			or ($self->_add_missing_path($hash) and link($source_file, $dest_name))
+			or rename("$source_file", $dest_name)
+			or ($self->path_for_hash($hash, 1) and rename($source_file, $dest_name))
 			or croak "rename($source_file => $dest_name): $!";
 		# record that we added a new hash, if stats enabled.
 		if ($flags->{stats}) {
@@ -322,84 +498,10 @@ sub _commit_file {
 }
 
 
-sub put_file {
-	my ($self, $file, $flags)= @_;
-
-	# Copied logic from superclass, because we might not get there
-	my $is_cas_file= ref $file && ref($file)->isa('DataStore::CAS::File');
-	if ($flags->{reuse_hash} && $is_cas_file) {
-		$flags->{known_hashes} ||= {};
-		$flags->{known_hashes}{ $file->store->digest }= $file->hash;
-	}
-
-	# Here is where we detect opportunity to perform optimized hard-linking
-	#  when copying to and from CAS implementations which are backed by
-	#  plain files.
-	if ($flags->{hardlink}) {
-		my $hardlink_source= 
-			($is_cas_file && $file->can('local_file') and length $file->local_file)? $file->local_file
-			: (ref $file && ref($file)->isa('Path::Class::File'))? "$file"
-			: (!ref $file)? $file
-			: undef;
-		if (defined $hardlink_source) {
-			# Try hard-linking it.  If fails, (i.e. cross-device) fall back to regular behavior
-			my $hash=
-				try { $self->_put_hardlink($file, $hardlink_source, $flags) }
-				catch { undef; };
-			return $hash if defined $hash;
-		}
-	}
-	# Else use the default implementation which opens and reads the file.
-	goto \&DataStore::CAS::put_file;
-}
-
-sub _put_hardlink {
-	my ($self, $file, $hardlink_source, $flags)= @_;
-
-	# If we know the hash, try linking directly to the final name.
-	my $hash= $flags->{known_hashes}{$self->digest};
-	if (defined $hash) {
-		$self->_commit_file($hardlink_source, $hash, $flags);
-		return $hash;
-	}
-
-	# If we don't know the hash, we first link to a temp file, to find out
-	# whether we can, and then calculate the hash, and then rename our link.
-	# This way we can fall back to regular behavior without double-reading
-	# the source file.
-	
-	# Use File::Temp to atomically get a unique filename, which we use as a prefix.
-	my $temp_file= File::Temp->new( TEMPLATE => 'temp-XXXXXXXX', DIR => $self->path );
-	my $temp_link= $temp_file."-lnk";
-	link( $hardlink_source, $temp_link )
-		or return undef;
-	
-	# success - we don't need to copy the file, just checksum it and rename.
-	# use try/catch so we can unlink our tempfile
-	return
-		try {
-			# Calculate hash
-			open( my $handle, '<:raw', $temp_link ) or die "open: $!";
-			my $digest= $self->_new_digest->addfile($handle);
-			$hash= $digest->hexdigest;
-			close $handle or die "close: $!";
-
-			# link to final correct name
-			$self->_commit_file($temp_link, $hash, $flags);
-			unlink($temp_link);
-			$hash;
-		}
-		catch {
-			unlink($temp_link);
-			undef;
-		};
-}
-
-
 sub validate {
 	my ($self, $hash)= @_;
 
-	my $path= $self->_path_for_hash($hash);
+	my $path= $self->path_for_hash($hash);
 	return undef unless -f $path;
 
 	open (my $fh, "<:raw", $path)
@@ -460,7 +562,7 @@ sub iterator {
 
 sub delete {
 	my ($self, $digest_hash, $flags)= @_;
-	my $path= $self->_path_for_hash($digest_hash);
+	my $path= $self->path_for_hash($digest_hash);
 	if (-f $path) {
 		unlink $path || die "unlink: $!"
 			unless $flags && $flags->{dry_run};
@@ -475,28 +577,22 @@ sub delete {
 }
 
 # This can be called as class or instance method.
-# When called as an instance method, '$digest_name' is mandatory,
+# When called as a class method, '$digest_name' is mandatory,
 #   otherwise it is unneeded.
 sub _new_digest {
 	my ($self, $digest_name)= @_;
 	Digest->new($digest_name || $self->digest);
 }
 
-sub _path_for_hash {
-	my ($self, $hash)= @_;
-	return catfile($self->path, ($hash =~ $self->_fanout_regex));
-}
-
-sub _add_missing_path {
-	my ($self, $hash)= @_;
-	my $str= $self->path;
-	my @parts= ($hash =~ $self->_fanout_regex);
-	pop @parts; # discard filename
-	for (@parts) {
-		$str= catdir($str, $_);
-		next if -d $str;
-		mkdir($str) or croak "mkdir($str): $!";
+sub _assert_digest_available {
+	my ($class, $digest)= @_;
+	try {
+		$class->_new_digest($digest)
 	}
+	catch {
+		s/^/# /mg;
+		croak "Digest algorithm '$digest' is not available on this system.\n$_\n"
+	};
 	1;
 }
 
@@ -515,13 +611,15 @@ __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 DataStore::CAS::Simple - Simple file/directory based CAS implementation
 
 =head1 VERSION
 
-version 0.020001
+version 0.03
 
 =head1 DESCRIPTION
 
@@ -562,14 +660,18 @@ set in the constructor when a new store is being created.  Default is C<SHA-1>.
 
 Read-only.  Returns arrayref of pattern used to split digest hashes into
 directories.  Each digit represents a number of characters from the front
-of the hash which then become a directory name.
+of the hash which then become a directory name.  The final digit may be
+the character '=' to indicate the filename is the full hash, or '*' to
+indicate the filename is the remaining digits of the hash.  '*' is the
+default behavior if the C<fanout> does not include one of these characters.
 
 For example, C<[ 2, 2 ]> would turn a hash of "1234567890" into a path of
-"12/34/567890".
+"12/34/567890".  C<[ 2, 2, '=' ]> would turn a hash of "1234567890" into
+a path of "12/34/1234567890".
 
 =head2 fanout_list
 
-Convenience accessor for C<@{ $cas-E<gt>fanout }>
+Convenience accessor for C<< @{ $cas->fanout } >>
 
 =head2 copy_buffer_size
 
@@ -582,12 +684,6 @@ Hashref of version information about the modules that created the store.
 Newer library versions can determine whether the storage is using an old
 format using this information.
 
-=head2 _fanout_regex
-
-Read-only.  A regex-ref which splits a digest hash into the parts needed
-for the path name.
-A fanout of C<[ 2, 2 ]> creates a regex of C</(.{2})(.{2})(.*)/>.
-
 =head1 METHODS
 
 =head2 new
@@ -616,6 +712,23 @@ newer version of the DataStore::CAS::Simple package that you are now using.
 
 =for Pod::Coverage BUILD
 
+=head2 path_parts_for_hash
+
+  my (@path)= $cas->path_parts_for_hash($digest_hash);
+
+Given a hash string, return the directory parts and filename where that content
+would be found.  They are returned as a list.  If the hash is not valid for this
+digest algorithm, this will throw an exception.
+
+=head2 path_for_hash
+
+  my $path= $cas->path_for_hash($digest_hash);
+  my $path= $cas->path_for_hash($digest_hash, $create_dirs);
+
+Given a hash string, return the path to the file, including C<< $self->path >>.
+The second argument can be set to true to create any missing directories in
+this path.
+
 =head2 create_store
 
   $class->create_store( %configuration | \%configuration )
@@ -634,14 +747,6 @@ this method using the relevant parameters you supplied to the constructor.
 
 See L<DataStore::CAS/get> for details.
 
-=head2 new_write_handle
-
-See L<DataStore::CAS/new_write_handle> for details.
-
-=head2 commit_write_handle
-
-See L<DataStore::CAS/commit_write_handle> for details.
-
 =head2 put
 
 See L<DataStore::CAS/put> for details.
@@ -655,10 +760,21 @@ See L<DataStore::CAS/put_scalar> for details.
 See L<DataStore::CAS/put_file> for details. In particular, heed the warnings
 about using the 'hardlink' and 'reuse_hash' flag.
 
-DataStore::CAS::Simple has special support for the flag 'hardlink'.  If your
-source is a real file, or instance of L<DataStore::CAS::File> from another
-DataStore::CAS::Simple, C<{ hardlink =E<gt> 1 }> will link to the file instead
-of copying it.
+DataStore::CAS::Simple has special support for the flags 'move' and 'hardlink'.
+If your source is a real file on the same filesystem by the same owner and/or
+group, C<< { move => 1 } >> will move the file instead of copying it.  (If it
+is a different filesystem or ownership can't be changed, it gets copied and the
+original gets unlinked).  If the file is a real file on the same filesystem with
+correct owner and permissions, C<< { hardlink => 1 } >> will link the file into
+the CAS instead of copying it.
+
+=head2 new_write_handle
+
+See L<DataStore::CAS/new_write_handle> for details.
+
+=head2 commit_write_handle
+
+See L<DataStore::CAS/commit_write_handle> for details.
 
 =head2 validate
 
@@ -696,7 +812,7 @@ Michael Conrad <mconrad@intellitree.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2013 by Michael Conrad, and IntelliTree Solutions llc.
+This software is copyright (c) 2022 by Michael Conrad, and IntelliTree Solutions llc.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
